//
// PS2PC: An Intel x86 emulator for MIPS R5900 (Sony PS2).
// Coded by Mike Tedder / breakpoint - bpoint@parodius.com
// http://ps2pc.parodius.com
//
// This project is released as 'open source'.  You may make any modifications
// or adjustments to the code and submit a patch to be applied to the project.
// See the web page for more details.
//
//-----------------------------------------------------------------------

#include "config.h"
#include "types.h"
#include "mem.h"
#include "debug.h"
#include "r59basic.h"

#include <stdlib.h>				// for malloc() and free()

void  *memBasePtr = NULL;			// Pointer to base of PC (x86) RAM...
void  *memAdjPtr  = NULL;			// "Adjusted" pointer to base of PC RAM (due to optimization trick).
ulong  memSize;						// Size of memory pool.
ulong  memLUIop;					// First portion of opcode to load base of RAM to AT.
ulong  memORIop;					// Second portion of opcode to load base of RAM to AT.


bool memInit(void)
{
	ulong dist;

	// Check we're not already inited...
	if (memBasePtr != NULL)
	{
		debugPrintf("memory system already inited!\n");
		return FALSE;
	}

	// Allocate RAM (plus 32k)...
	memSize    = CFG_X86_MEMORY_SIZE * 1024 * 1024;
	memBasePtr = malloc(memSize + (32 * 1024));
	if (memBasePtr == NULL) return FALSE;

	debugPrintf("memory base inited at address [%08X]\n", memBasePtr);

	// Confirm our base pointer is < 8000h as MIPS memory opcodes sign extend the immediate.  If it's over, adjust the base.
	if ((ushort)memBasePtr >= 0x8000)
	{
		dist = 0x10000 - (ushort)memBasePtr;
		memAdjPtr = (void *) ((uchar *)memBasePtr + dist);
		debugPrintf("memory base adjusted to compensate for optimization.\n");
	}
	else
	{
		// No adjustment necessary...
		memAdjPtr = memBasePtr;
		debugPrintf("no adjustment necessary to base memory pointer.\n");
	}

	// Create two opcodes to allow AT to point to base of PC RAM.
	memLUIop = r5900LUI(rAT,      (ushort) (((ulong)memAdjPtr) >> 16) );
	memORIop = r5900ORI(rAT, rR0, (ushort) (        memAdjPtr)        );

	// All done!
	return TRUE;
}

void memShutdown(void)
{
	// Make sure we're not already shutdown...
	if (memBasePtr == NULL)
	{
		debugPrintf("memory system already shutdown!\n");
		return;
	}

	free(memBasePtr);
	memBasePtr = NULL;
}

void *memGetBase(void)
{
	return memAdjPtr;
}

ulong memGetSize(void)
{
	return memSize;
}

void *memCalcSegOfsAddr(ushort seg, ushort ofs)
{
	return ( ((uchar *)memAdjPtr) + (seg << 4) + ofs );
}


ulong mem16WriteByte(ulong *out, x86RegByte srcReg, x86MODRMByte16 destAddr, ushort disp)
{
	x86Seg currSeg;
	uint size = 0;

	// Create an offset based on addressing mode.
	switch (destAddr)
	{
		case ea16DS_BXpSI:
		case ea16DS_BXpSIpd8:
		case ea16DS_BXpSIpd16:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			* out      = r5900ANDI(rS0, rT1, 0xFFFF);			// take lower 16 bits of EBX
			*(out + 1) = r5900ANDI(rS1, rT4, 0xFFFF);			// take lower 16 bits of ESI
			*(out + 2) = r5900ADD (rAT, rS0, rS1);				// add them
			size       = 3;
			break;

		case ea16DS_BXpDI:
		case ea16DS_BXpDIpd8:
		case ea16DS_BXpDIpd16:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			* out      = r5900ANDI(rS0, rT1, 0xFFFF);			// take lower 16 bits of EBX
			*(out + 1) = r5900ANDI(rS1, rT5, 0xFFFF);			// take lower 16 bits of EDI
			*(out + 2) = r5900ADD (rAT, rS0, rS1);				// add them
			size       = 3;
			break;

		case ea16SS_BPpSI:
		case ea16SS_BPpSIpd8:
		case ea16SS_BPpSIpd16:
			// Initialize default segment...
			currSeg = xSS;

			// Begin encoding.
			* out      = r5900ANDI(rS0, rT6, 0xFFFF);			// take lower 16 bits of EBP
			*(out + 1) = r5900ANDI(rS1, rT4, 0xFFFF);			// take lower 16 bits of ESI
			*(out + 2) = r5900ADD (rAT, rS0, rS1);				// add them
			size       = 3;
			break;

		case ea16SS_BPpDI:
		case ea16SS_BPpDIpd8:
		case ea16SS_BPpDIpd16:
			// Initialize default segment...
			currSeg = xSS;

			// Begin encoding.
			* out      = r5900ANDI(rS0, rT6, 0xFFFF);			// take lower 16 bits of EBP
			*(out + 1) = r5900ANDI(rS1, rT5, 0xFFFF);			// take lower 16 bits of EDI
			*(out + 2) = r5900ADD (rAT, rS0, rS1);				// add them
			size       = 3;
			break;

		case ea16DS_SI:
		case ea16DS_SIpd8:
		case ea16DS_SIpd16:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900ANDI(rAT, rT4, 0xFFFF);					// take lower 16 bits of ESI
			size = 1;
			break;

		case ea16DS_DI:
		case ea16DS_DIpd8:
		case ea16DS_DIpd16:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900ANDI(rAT, rT5, 0xFFFF);					// take lower 16 bits of EDI
			size = 1;
			break;

		case ea16SS_BPpd8:
		case ea16SS_BPpd16:
			// Initialize default segment...
			currSeg = xSS;

			// Begin encoding.
			*out = r5900ANDI(rAT, rT6, 0xFFFF);					// take lower 16 bits of EBP
			size = 1;
			break;

		case ea16DS_BX:
		case ea16DS_BXpd8:
		case ea16DS_BXpd16:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900ANDI(rAT, rT1, 0xFFFF);					// take lower 16 bits of EBX
			size = 1;
			break;

		case ea16DS_d16:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rR0);						// no base register
			size = 1;
			break;

		default:
			debugPrintf("[mem16WriteByte] Unknown addressing mode (%02Xh)\n", destAddr);
			return 0;
	}

	// Check for displacement...
	if (disp)
	{
		// Manually sign extend the displacement if it is 8-bit.  (ADDI auto sign-extends a 16-bit immediate.)
		if (MODRM_HAS_DISPL8(destAddr) && (disp & 0x80))
			disp |= 0xFF00;

		*(out + (size++)) = r5900ADDI(rS0, rAT, disp);		// add in the displacement
		*(out + (size++)) = r5900ANDI(rS1, rS0, 0xFFFF);	// chop off upper portion (in case of rollover)
	}
	else
	{
		*(out + (size++)) = r5900ANDI(rS1, rAT, 0xFFFF);	// again, chop off upper portion
	}

	// Handle any segment overrides...
	if (gIsSegOverride) currSeg = gOverrideSeg;

	// Factor in the segment with the offset and create a physical address.
	*(out + (size++)) = r5900ORI(rAT, rR0, x86SegToR5900Reg[currSeg]);
	*(out + (size++)) = r5900SLL(rS0, rAT, 4);
	*(out + (size++)) = r5900ADD(rS0, rS0, rS1);			// add in calced offset from above (S1)

	// Initialize the upper half of AT with base of x86 RAM.  (The lower half is used in the 'offset' of SB.)
	*(out + (size++)) = memLUIop;

	// Now add the x86 physical address with the base of RAM.
	*(out + (size++)) = r5900ADD(rS1, rAT, rS0);

	// Create a 'store byte' opcode writing the appropriate register to memory...
	switch (srcReg)
	{
		case xAL: *(out + (size++)) = r5900SB(rT0, rS1, (ushort)memAdjPtr); break;
		case xBL: *(out + (size++)) = r5900SB(rT1, rS1, (ushort)memAdjPtr); break;
		case xCL: *(out + (size++)) = r5900SB(rT2, rS1, (ushort)memAdjPtr); break;
		case xDL: *(out + (size++)) = r5900SB(rT3, rS1, (ushort)memAdjPtr); break;

		case xAH:
			*(out + (size++)) = r5900SRL(rS0, rT0, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xBH:
			*(out + (size++)) = r5900SRL(rS0, rT1, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xCH:
			*(out + (size++)) = r5900SRL(rS0, rT2, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xDH:
			*(out + (size++)) = r5900SRL(rS0, rT3, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;
	}

	return size;
}

ulong mem32WriteByte(ulong *out, x86RegByte srcReg, x86MODRMByte32 destAddr, ulong disp)
{
	x86Seg currSeg;
	uint size = 0;

	// Create an offset based on addressing mode.
	switch (destAddr)
	{
		case ea32DS_EAX:
		case ea32DS_EAXpd8:
		case ea32DS_EAXpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT0);						// copy over EAX
			size = 1;
			break;

		case ea32DS_ECX:
		case ea32DS_ECXpd8:
		case ea32DS_ECXpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT2);						// copy over ECX
			size = 1;
			break;

		case ea32DS_EDX:
		case ea32DS_EDXpd8:
		case ea32DS_EDXpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT3);						// copy over EDX
			size = 1;
			break;

		case ea32DS_EBX:
		case ea32DS_EBXpd8:
		case ea32DS_EBXpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT1);						// copy over EBX
			size = 1;
			break;

		case ea32SS_EBPpd8:
		case ea32SS_EBPpd32:
			// Initialize default segment...
			currSeg = xSS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT6);						// copy over EBP
			size = 1;
			break;

		case ea32DS_ESI:
		case ea32DS_ESIpd8:
		case ea32DS_ESIpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT4);						// copy over ESI
			size = 1;
			break;

		case ea32DS_EDI:
		case ea32DS_EDIpd8:
		case ea32DS_EDIpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT5);						// copy over EDI
			size = 1;
			break;

		case ea32DS_d32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rR0);						// no base register
			size = 1;
			break;

		default:
			debugPrintf("[mem32WriteByte] Unknown addressing mode (%02Xh)\n", destAddr);
			return 0;
	}

	// Check for displacement...
	if (disp)
	{
		// Manually sign extend the displacement if it is 8-bit.
		if (MODRM_HAS_DISPL8(destAddr) && (disp & 0x80))
			disp |= 0xFFFFFF00;

		*(out + (size++)) = r5900LUI(rS0, HI_WORD(disp));	// create 32-bit displacement
		*(out + (size++)) = r5900ORI(rS0, rS0, LO_WORD(disp));
		*(out + (size++)) = r5900ADD(rS1, rAT, rS0);		// and add it in...
	}
	else
	{
		*(out + (size++)) = r5900OR(rS1, rAT, rR0);			// carry over above value since there's no displacement.
	}

	// Handle any segment overrides...
	if (gIsSegOverride) currSeg = gOverrideSeg;

	// Factor in the segment with the offset and create a physical address.
	*(out + (size++)) = r5900ORI(rAT, rR0, x86SegToR5900Reg[currSeg]);
	*(out + (size++)) = r5900SLL(rS0, rAT, 4);
	*(out + (size++)) = r5900ADD(rS0, rS0, rS1);			// add in calced offset from above (S1)

	// Initialize the upper half of AT with base of x86 RAM.  (The lower half is used in the 'offset' of SB.)
	*(out + (size++)) = memLUIop;

	// Now add the x86 physical address with the base of RAM.
	*(out + (size++)) = r5900ADD(rS1, rAT, rS0);

	// Create a 'store byte' opcode writing the appropriate register to memory...
	switch (srcReg)
	{
		case xAL: *(out + (size++)) = r5900SB(rT0, rS1, (ushort)memAdjPtr); break;
		case xBL: *(out + (size++)) = r5900SB(rT1, rS1, (ushort)memAdjPtr); break;
		case xCL: *(out + (size++)) = r5900SB(rT2, rS1, (ushort)memAdjPtr); break;
		case xDL: *(out + (size++)) = r5900SB(rT3, rS1, (ushort)memAdjPtr); break;

		case xAH:
			*(out + (size++)) = r5900SRL(rS0, rT0, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xBH:
			*(out + (size++)) = r5900SRL(rS0, rT1, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xCH:
			*(out + (size++)) = r5900SRL(rS0, rT2, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xDH:
			*(out + (size++)) = r5900SRL(rS0, rT3, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;
	}

	return size;
}

ulong mem32SIBWriteByte(ulong *out, x86RegByte srcReg, x86MODRMByte32SIB sibAddr, uchar ss, uchar index, ulong disp)
{
	x86Seg currSeg;
	uint size = 0;

	// Create an offset based on addressing mode.
	switch (sibAddr)
	{
		case ea32sibDS_EAX:
		case ea32sibDS_EAXpd8:
		case ea32sibDS_EAXpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT0);						// copy over EAX
			size = 1;
			break;

		case ea32sibDS_ECX:
		case ea32sibDS_ECXpd8:
		case ea32sibDS_ECXpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT2);						// copy over ECX
			size = 1;
			break;

		case ea32sibDS_EDX:
		case ea32sibDS_EDXpd8:
		case ea32sibDS_EDXpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT3);						// copy over EDX
			size = 1;
			break;

		case ea32sibDS_EBX:
		case ea32sibDS_EBXpd8:
		case ea32sibDS_EBXpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT1);						// copy over EBX
			size = 1;
			break;

		case ea32sibSS_ESP:
		case ea32sibSS_ESPpd8:
		case ea32sibSS_ESPpd32:
			// Initialize default segment...
			currSeg = xSS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT7);						// copy over ESP
			size = 1;
			break;

		case ea32sibSS_EBPpd8:
		case ea32sibSS_EBPpd32:
			// Initialize default segment...
			currSeg = xSS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT6);						// copy over EBP
			size = 1;
			break;

		case ea32sibDS_ESI:
		case ea32sibDS_ESIpd8:
		case ea32sibDS_ESIpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT4);						// copy over ESI
			size = 1;
			break;

		case ea32sibDS_EDI:
		case ea32sibDS_EDIpd8:
		case ea32sibDS_EDIpd32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rT5);						// copy over EDI
			size = 1;
			break;

		case ea32sibDS_d32:
			// Initialize default segment...
			currSeg = xDS;

			// Begin encoding.
			*out = r5900OR(rAT, rR0, rR0);						// no base register
			size = 1;
			break;

		default:
			debugPrintf("[mem32SIBWriteByte] Unknown addressing mode (%02Xh)\n", sibAddr);
			return 0;
	}

	// Check for displacement...
	if (disp)
	{
		// Manually sign extend the displacement if it is 8-bit.
		if (SIB_HAS_DISPL8(sibAddr) && (disp & 0x80))
			disp |= 0xFFFFFF00;

		*(out + (size++)) = r5900LUI(rS0, HI_WORD(disp));	// create 32-bit displacement
		*(out + (size++)) = r5900ORI(rS0, rS0, LO_WORD(disp));
		*(out + (size++)) = r5900ADD(rS1, rAT, rS0);		// and add it in...
	}
	else
	{
		*(out + (size++)) = r5900OR(rS1, rAT, rR0);			// carry over above value since there's no displacement.
	}

	// Handle scaled index register...
	if (index != 0x04)
	{
		// Create scalar value.
		switch (index)
		{
			case 0x00: *(out + (size++)) = r5900SLL(rAT, rT0, ss); break;		// EAX
			case 0x01: *(out + (size++)) = r5900SLL(rAT, rT2, ss); break;		// ECX
			case 0x02: *(out + (size++)) = r5900SLL(rAT, rT3, ss); break;		// EDX
			case 0x03: *(out + (size++)) = r5900SLL(rAT, rT1, ss); break;		// EBX
			//no 0x04...
			case 0x05: *(out + (size++)) = r5900SLL(rAT, rT6, ss); break;		// EBP
			case 0x06: *(out + (size++)) = r5900SLL(rAT, rT4, ss); break;		// ESI
			case 0x07: *(out + (size++)) = r5900SLL(rAT, rT5, ss); break;		// EDI
		}

		// Add it into the current offset.
		*(out + (size++)) = r5900ADD(rS1, rS1, rAT);
	}

	// Handle any segment overrides...
	if (gIsSegOverride) currSeg = gOverrideSeg;

	// Factor in the segment with the offset and create a physical address.
	*(out + (size++)) = r5900ORI(rAT, rR0, x86SegToR5900Reg[currSeg]);
	*(out + (size++)) = r5900SLL(rS0, rAT, 4);
	*(out + (size++)) = r5900ADD(rS0, rS0, rS1);			// add in calced offset from above (S1)

	// Initialize the upper half of AT with base of x86 RAM.  (The lower half is used in the 'offset' of SB.)
	*(out + (size++)) = memLUIop;

	// Now add the x86 physical address with the base of RAM.
	*(out + (size++)) = r5900ADD(rS1, rAT, rS0);

	// Create a 'store byte' opcode writing the appropriate register to memory...
	switch (srcReg)
	{
		case xAL: *(out + (size++)) = r5900SB(rT0, rS1, (ushort)memAdjPtr); break;
		case xBL: *(out + (size++)) = r5900SB(rT1, rS1, (ushort)memAdjPtr); break;
		case xCL: *(out + (size++)) = r5900SB(rT2, rS1, (ushort)memAdjPtr); break;
		case xDL: *(out + (size++)) = r5900SB(rT3, rS1, (ushort)memAdjPtr); break;

		case xAH:
			*(out + (size++)) = r5900SRL(rS0, rT0, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xBH:
			*(out + (size++)) = r5900SRL(rS0, rT1, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xCH:
			*(out + (size++)) = r5900SRL(rS0, rT2, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;

		case xDH:
			*(out + (size++)) = r5900SRL(rS0, rT3, 8);
			*(out + (size++)) = r5900SB (rS0, rS1, (ushort)memAdjPtr);
			break;
	}

	return size;
}
